﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.IO;
using System.Text;

namespace P2PQRelay
{
    public class Discover
    {
        public static IPEndPoint Query()
        {
            int qport = P2PQRelayClient_Net.Properties.Settings.Default.QueryPort;

            IPEndPoint target_ep = null;

            Socket recvSocket = null;
            Socket sendSocket = null;

            try
            {
                // ----- 応答受信用ソケットをリッスン
                recvSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                recvSocket.DontFragment = true;
                recvSocket.EnableBroadcast = true;
                recvSocket.MulticastLoopback = true;
                EndPoint recv_ep = new IPEndPoint(IPAddress.Any, 0);
                recvSocket.Bind(recv_ep);
                int resp_port = ((IPEndPoint)(recvSocket.LocalEndPoint)).Port;

                // ----- 応答受信開始(非同期)
                byte[] recvBuf = new byte[1024];
                EndPoint ep = new IPEndPoint(0, 0);
                ManualResetEvent received = new ManualResetEvent(false);
                IAsyncResult receiving = recvSocket.BeginReceiveFrom(recvBuf, 0, recvBuf.Length, SocketFlags.None, ref ep, (AsyncCallback)(ar => {
                    ManualResetEvent signal = ar.AsyncState as ManualResetEvent;
                    signal.Set();
                }), received);

                // ----- 要求を送信
                byte[] sendData = Encoding.UTF8.GetBytes("P2PQRelay/1.0 DiscoveryServer Query " + resp_port.ToString());

                sendSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
                sendSocket.DontFragment = true;
                sendSocket.EnableBroadcast = true;
                sendSocket.MulticastLoopback = true;
                EndPoint send_ep = new IPEndPoint(IPAddress.Parse("255.255.255.255"), qport);
                sendSocket.SendTo(sendData, sendData.Length, SocketFlags.None, send_ep);
                sendSocket.Close(); sendSocket = null;

                // ----- 応答受信待ち(1000msでタイムアウト)
                if (received.WaitOne(1000))
                {
                    EndPoint remote_ep = new IPEndPoint(0, 0);
                    int size = recvSocket.EndReceiveFrom(receiving, ref remote_ep);
                    System.Diagnostics.Debug.WriteLine("received " + Encoding.UTF8.GetString(recvBuf, 0, size));

                    const string RECV_OK = "P2PQRelay/1.0 DiscoveryServer OK ";
                    string recvstr = Encoding.UTF8.GetString(recvBuf, 0, size);
                    if (recvstr.StartsWith(RECV_OK))
                    {
                        string[] ipport = recvstr.Substring(RECV_OK.Length).Split(':');

                        target_ep = new IPEndPoint(IPAddress.Parse(ipport[0]), int.Parse(ipport[1]));
                    }
                }

                recvSocket.Close(); recvSocket = null;
            }
            catch (Exception e)
            {
                System.Diagnostics.Debug.WriteLine(e);
            }
            finally
            {
                if (null != sendSocket)
                {
                    sendSocket.Close();
                }
                if (null != recvSocket)
                {
                    recvSocket.Close();
                }
            }

            return target_ep;
        }
    }
}
